import json
import logging
import os
import sqlite3
from typing import List

import numpy as np
import pandas as pd
from qdrant_client import QdrantClient
from transformers import AutoModel

from agent_search.core import SERPResult
from agent_search.core.utils import cosine_similarity, load_config

logger = logging.getLogger(__name__)


class OpenWebSearch:
    """A simple search client for the OpenSearch collection"""

    def __init__(
        self,
    ):
        # Load the configuration
        self.config = load_config()["open_web_search"]
        logger.info(
            f"Connecting to collection: {self.config['qdrant_collection_name']}"
        )
        self.collection_name = self.config["qdrant_collection_name"]
        self.client = QdrantClient(
            self.config["qdrant_client_host"],
            grpc_port=self.config["qdrant_client_grpc_port"],
            prefer_grpc=True,
        )

        self.embedding_model = AutoModel.from_pretrained(
            self.config["embedding_model_name"], trust_remote_code=True
        )

        self.sqlite_table_name = self.config["sqlite_table_name"]

        self.pagerank_rerank_module = self.config["pagerank_rerank_module"]
        pagerank_file_path = self.config["pagerank_file_path"]

        if (
            not os.path.exists(pagerank_file_path)
            and self.pagerank_rerank_module
        ):
            raise ValueError(
                "Must have pagerank_file_path when using pagerank_rerank_module"
            )

        if self.pagerank_rerank_module:
            if not pagerank_file_path:
                # Simulating reading from a CSV file
                pagerank_file_path = os.path.join(
                    os.path.dirname(__file__), "..", "data", "domain_ranks.csv"
                )

            # Reading the CSV data using pandas
            df = pd.read_csv(pagerank_file_path)
            self.domain_to_rank_map = dict(
                zip(df["Domain"], df["Open Page Rank"])
            )
            self.pagerank_rerank_module = True
            self.pagerank_importance = float(
                self.config["pagerank_importance"]
            )

    def get_query_vector(self, query: str):
        """Gets the query vector for the given query"""

        query_vector = self.embedding_model.encode(query)
        return query_vector

    def similarity_search(
        self,
        query_vector: np.ndarray,
        limit: int = 100,
    ):
        """Searches the collection for the given query and returns the top 'limit' results"""

        points = self.client.search(
            collection_name=self.collection_name,
            query_vector=query_vector,
            limit=limit,
        )

        return [
            SERPResult(
                score=point.score,
                text=point.payload["text"],
                title=None,
                url=point.payload["url"],
                metadata={},
            )
            for point in points
        ]

    def hierarchical_similarity_reranking(
        self,
        query_vector: np.ndarray,
        urls: List[str],
        limit: int = 100,
    ) -> List[SERPResult]:
        """Hierarchical URL search to find the most similar text chunk for the given query and URLs"""

        conn = sqlite3.connect(self.config["sqlite_db_rel_path"])
        cur = conn.cursor()

        # SQL query to fetch the entries for the URLs
        # Assuming 'urls' is a list of URL strings
        placeholders = ", ".join("?" * len(urls))
        query = f"SELECT * FROM {self.sqlite_table_name} WHERE url IN ({placeholders})"

        # Fetch all results
        cur.execute(query, tuple(urls))
        results = cur.fetchall()

        # List to store the results along with their similarity scores
        similarity_results = []

        # Iterate over each result to find the most similar text chunk
        for result in results:
            (
                _,
                url,
                title,
                metadata,
                dataset,
                text_chunks_str,
                embeddings_str,
            ) = result
            text_chunks = json.loads(text_chunks_str)
            embeddings = json.loads(embeddings_str)
            max_similarity = -1
            most_similar_chunk = None

            # Iterate over each embedding to find the one with maximum cosine similarity
            for chunk, embedding in zip(text_chunks, embeddings):
                similarity = cosine_similarity(
                    np.array(query_vector), np.array(embedding)
                )
                if similarity > max_similarity:
                    max_similarity = similarity
                    most_similar_chunk = chunk

            # Store the most similar chunk and its similarity score
            similarity_results.append(
                SERPResult(
                    score=max_similarity,
                    url=url,
                    title=title,
                    metadata=json.loads(metadata),
                    dataset=dataset,
                    text=most_similar_chunk,
                ),
            )

        # Sort the results based on similarity score in descending order
        similarity_results.sort(key=lambda x: x.score, reverse=True)
        conn.close()
        return similarity_results[:limit]

    def pagerank_reranking(
        self,
        similarity_results: List[SERPResult],
        limit: int = 100,
    ) -> List[SERPResult]:
        """Reranks the results based on the PageRank score of the domain"""
        if not self.pagerank_rerank_module:
            raise Exception(
                "PageRank reranking module is not enabled. Please set pagerank_rerank_module=True while initializing the OpenWebSearch client."
            )
        # List to store the results along with their PageRank scores
        pagerank_results = []

        # Iterate over each result to find the PageRank score of the domain
        for result in similarity_results:
            pagerank_score = 0
            try:
                domain = result.url.split("/")[2]
                pagerank_score = self.domain_to_rank_map.get(domain, 0)
            except Exception as e:
                logger.info(f"Error {e}: Found for URL: {result.url}")
            reweighted_score = (
                self.pagerank_importance * pagerank_score / 10.0
                + (1 - self.pagerank_importance) * result.score
            )
            pagerank_results.append(
                SERPResult(
                    score=reweighted_score,
                    url=result.url,
                    title=result.title,
                    metadata=result.metadata,
                    dataset=result.dataset,
                    text=result.text,
                )
            )

        # Sort the results based on PageRank score in descending order
        pagerank_results.sort(key=lambda x: x.score, reverse=True)
        return pagerank_results[:limit]



import fire
import logging

import requests

from agent_search.core import SERPClient

logger = logging.getLogger(__name__)


class QueryWrapper:
    """A wrapper class to run queries on the SERPClient"""

    def __init__(self):
        pass

    def query(
        self, query="What is a lagrangian?", api_base="http://0.0.0.0:8000"
    ):
        """Run a query on the SERPClient"""
        client = SERPClient(api_base)
        logging.basicConfig(level=logging.INFO)

        try:
            results = client.search(query)
            for i, result in enumerate(results):
                logging.info(
                    f"{i}. \033[94mURL: {result.url}\033[0m (Score: \033[95m{result.score:.2f}\033[0m)"
                )
                logging.info("-" * 50)
                logging.info(f"Title: \033[93m{result.title}\033[0m")
                logging.info(f"Text:\n{result.text}\n")
                # logging.info(f"Metadata:\n{result.metadata}...")
                logging.info("-" * 80)

        except requests.HTTPError as e:
            logging.info(f"An error occurred: {e}")


if __name__ == "__main__":
    fire.Fire(QueryWrapper)



from typing import Any, Optional

from pydantic import BaseModel


class SERPResult(BaseModel):
    """A dataclass to store the search result"""

    score: float
    url: str
    title: Optional[str]
    dataset: Optional[str]
    # TODO - Add dict(str, [str, float, ..]) validation
    metadata: Any
    text: str

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.title and self.title == self.text[0 : len(self.title)]:
            self.text = self.text[len(self.title) :]
        self.text = self.text.strip()



from typing import List

import requests


class SERPResult:
    def __init__(
        self,
        score: float,
        url: str,
        title: str,
        dataset: str,
        metadata: dict,
        text: str,
    ):
        self.score = score
        self.url = url
        self.title = title
        self.dataset = dataset
        self.metadata = metadata
        self.text = text.strip()
        if self.title and self.text.startswith(self.title):
            self.text = self.text[len(self.title) :].strip()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


class SERPClient:
    def __init__(self, api_base: str):
        self.api_base = api_base

    def search(
        self,
        query: str,
        num_step0_results: int = 100,
        num_step1_results: int = 25,
        num_step2_results: int = 10,
    ) -> List[SERPResult]:
        payload = {
            "query": query,
            "num_step0_results": num_step0_results,
            "num_step1_results": num_step1_results,
            "num_step2_results": num_step2_results,
        }
        response = requests.post(f"{self.api_base}/search", json=payload)
        response.raise_for_status()  # Raises an HTTPError if the HTTP request returned an unsuccessful status code

        results = response.json().get("results", [])
        return [SERPResult.from_dict(result) for result in results]



"""A script to populate the database with the given dataset and subset."""
import json
import logging
import os
import queue
import sqlite3
import threading
import uuid

import fire
from datasets import load_dataset
from qdrant_client import QdrantClient
from qdrant_client.http import models
from qdrant_client.http.models import PointStruct

from agent_search.core.utils import load_config

logger = logging.getLogger(__name__)


# A function that will write out the points
# To be ran in a separate thread
def writer(o, config, qclient):
    logger.info("Starting writer...")
    i = 0
    while True:
        try:
            points = o.get()
            if points is None:
                break
            logger.info(f"Writing out...")
            operation_info = qclient.upsert(
                collection_name=config["qdrant_collection_name"],
                wait=True,
                points=points,
            )
            i += len(points)
            logger.info(
                f"i:{i}, n_points:{len(points)}, operation_info:{operation_info}"
            )

            o.task_done()
        except Exception as e:
            logger.info(f"Error: {e}")
            continue


def create_sqlite_table(conn, cursor, table_name) -> None:
    """Create the table in the database"""
    create_sqlite_table_query = f"""
    CREATE TABLE IF NOT EXISTS {table_name} (
        id INTEGER PRIMARY KEY,
        url TEXT UNIQUE,
        title TEXT,
        metadata TEXT,
        dataset TEXT,
        text_chunks TEXT,
        embeddings TEXT
    )
    """
    cursor.execute(create_sqlite_table_query)
    conn.commit()


def create_qdrant_collection(client, config) -> None:
    client.create_collection(
        collection_name=config["qdrant_collection_name"],
        vectors_config=models.VectorParams(
            size=768, distance=models.Distance.COSINE
        ),
        quantization_config=models.ScalarQuantization(
            scalar=models.ScalarQuantizationConfig(
                type=models.ScalarType.INT8,
                quantile=0.99,
                always_ram=True,
            ),
        ),
    )


def insert_entry(conn, cursor, table_name, entry):
    """Insert the entry into the database"""

    # Serialize the embeddings list to a JSON string
    text_chunks_json = json.dumps(entry["text_chunks"])
    embeddings_json = json.dumps(entry["embeddings"])

    insert_query = f"""
    INSERT INTO {table_name} (id, url, title, metadata, dataset, text_chunks, embeddings) 
    VALUES (?, ?, ?, ?, ?, ?, ?)
    """
    try:
        cursor.execute(
            insert_query,
            (
                entry["id"],
                entry["url"],
                entry["title"],
                entry["metadata"],
                entry["dataset"],
                text_chunks_json,
                embeddings_json,
            ),
        )
        conn.commit()
    except sqlite3.IntegrityError as e:
        logger.warning(f"Error while inserting entry: {e}")


class Populate:
    """A class to populate the database with the given dataset and subset."""

    def __init__(self):
        self.config = load_config()["open_web_search"]

    def populate_sqlite(
        self,
        dataset="AgentSearch/OpenWebSearch-V1",
        subset="arxiv",
        limit=10_000,
        log_interval=1_000,
    ) -> None:
        """Populate the database with the given dataset and subset"""
        if subset == "all":
            subset = "**"

        db_path = Populate.get_db_path(self.config)

        if not os.path.exists(os.path.dirname(db_path)):
            os.makedirs(os.path.dirname(db_path))

        conn = sqlite3.connect(db_path)
        cur = conn.cursor()

        sqlite_table_name = self.config["sqlite_table_name"]
        create_sqlite_table(conn, cur, sqlite_table_name)

        for counter, entry in enumerate(
            Populate.data_streamer(dataset, subset)
        ):
            insert_entry(conn, cur, sqlite_table_name, entry)
            if counter % log_interval == 0:
                logger.info(f"Inserted {counter} entries")
            if counter >= limit:
                break

        # Count the number of rows in the table
        cur.execute(f"SELECT COUNT(*) FROM {sqlite_table_name}")
        row_count = cur.fetchone()[0]
        logger.info(f"Number of rows in the table: {row_count}")

        # Calculate the size of the database file
        db_size = os.path.getsize(db_path)
        logger.info(f"Size of the database file: {db_size} bytes")

        # Close the database connection
        conn.close()

    def populate_qdrant(self, batch_size=100) -> None:
        """Populate the database with the given dataset and subset"""

        output_queue = queue.Queue()

        qdrant_client = QdrantClient(
            self.config["qdrant_client_host"],
            grpc_port=self.config["qdrant_client_grpc_port"],
            prefer_grpc=True,
        )

        writer_thread = threading.Thread(
            target=writer,
            args=(output_queue, self.config, qdrant_client),
        )
        writer_thread.start()

        db_path = Populate.get_db_path(self.config)
        conn = sqlite3.connect(db_path)
        cur = conn.cursor()

        try:
            create_qdrant_collection(qdrant_client, self.config)
        except Exception as e:
            logger.warning(f"Error while creating collection: {e}")

        # Prepare the query to select entries from the SQLite database
        sqlite_table_name = self.config["sqlite_table_name"]
        select_query = f"SELECT * FROM {sqlite_table_name}"
        cur.execute(select_query)

        logger.info("Populating the Qdrant Collection...")
        while True:
            logger.info("Fetching rows from the SQLite database...")
            rows = cur.fetchmany(batch_size)

            if not rows:
                break

            # Process each row in the batch
            points = []
            for row in rows:
                # Extract and process data from the row
                (
                    _,
                    url,
                    __,
                    ___,
                    ____,
                    text_chunks,
                    embeddings,
                ) = row
                text_chunks_list = json.loads(text_chunks)
                embeddings_list = json.loads(embeddings)
                points.extend(
                    [
                        PointStruct(
                            id=str(uuid.uuid3(uuid.NAMESPACE_DNS, url)),
                            vector=embedding,
                            payload={"text": text_chunks_list[i], "url": url},
                        )
                        for i, embedding in enumerate(embeddings_list)
                    ]
                )

            output_queue.put(points)

            logger.info(f"Processed batch of {len(rows)} entries")

        output_queue.put(points)

        # Close the database connection
        conn.close()

    @staticmethod
    def get_db_path(config):
        return os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            config["sqlite_db_rel_path"],
        )

    @staticmethod
    def data_streamer(dataset, subset):
        """Stream the data from the given dataset and subset"""
        return load_dataset(
            dataset,
            data_files=f"{subset}/*",
            streaming=True,
        )["train"]


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logger.setLevel(logging.INFO)
    fire.Fire(Populate)



import logging
import time
from typing import Optional

import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel

from agent_search.core.utils import load_config, select_top_urls
from agent_search.search import OpenWebSearch

logger = logging.getLogger(__name__)


def timed_fn(fn):
    """A decorator to time a function"""

    def wrapper(*args, **kwargs):
        t0 = time.time()
        result = fn(*args, **kwargs)
        t1 = time.time()
        logger.debug(f"Time taken for {fn.__name__}: {t1 - t0}")
        return result

    return wrapper


class SearchServer:
    def __init__(self):
        self.client = OpenWebSearch()

    def run(
        self,
        query="What is a lagrangian?",
        limit_broad_results=1_000,
        limit_deduped_url_results=50,
        limit_hierarchical_url_results=20,
        limit_final_pagerank_results=10,
        url_contains_filter=None,
    ):
        """Run a search query using the OpenWebSearch client"""
        query_vector = self.client.get_query_vector(query)
        broad_results = timed_fn(self.client.similarity_search)(
            query_vector=query_vector, limit=limit_broad_results
        )
        logger.debug(
            f"Step 0: Broad similarity search, {len(broad_results)} results"
        )

        if not url_contains_filter:
            url_contains_filter = []

        deduped_url_results = select_top_urls(
            broad_results,
            max_urls=limit_deduped_url_results,
            url_contains=url_contains_filter,
        )

        logger.debug(
            f"Step 1: Unique URL Filtration, {len(deduped_url_results)} results"
        )

        hierarchical_url_results = timed_fn(
            self.client.hierarchical_similarity_reranking
        )(
            query_vector=query_vector,
            urls=deduped_url_results,
            limit=limit_hierarchical_url_results,
        )

        logger.debug(
            f"Step 2: Reranking using hierarchical similarity search, {len(hierarchical_url_results)} results returned"
        )

        try:
            pagerank_reranked_results = timed_fn(
                self.client.pagerank_reranking
            )(hierarchical_url_results)[:limit_final_pagerank_results]
            logger.debug(
                "Step 3: Reranking using pagerank, {len(pagerank_reranked_results)} results returned"
            )

            # Print or process the sorted results
            for serp_result in pagerank_reranked_results:
                logger.debug(
                    f"URL: {serp_result.url}, Similarity: {serp_result.score:.4f}:\nTitle: {serp_result.title}\nText:\n{serp_result.text}"
                )
                logger.debug("-" * 100)
        except Exception as e:
            logger.error(
                f"An error occurred while reranking using pagerank: {e}"
            )

        return pagerank_reranked_results


class SearchQuery(BaseModel):
    """A search query data model"""

    query: str
    limit_broad_results: Optional[int] = 1_000
    limit_deduped_url_results: Optional[int] = 100
    limit_hierarchical_url_results: Optional[int] = 25
    limit_final_pagerank_results: Optional[int] = 10


app = FastAPI()
search_runner = SearchServer()


@app.post("/search")
def run_search(query: SearchQuery):
    """Run a search query"""
    try:
        results = search_runner.run(
            query=query.query,
            limit_broad_results=query.limit_broad_results,
            limit_deduped_url_results=query.limit_deduped_url_results,
            limit_hierarchical_url_results=query.limit_hierarchical_url_results,
            limit_final_pagerank_results=query.limit_final_pagerank_results,
        )
        return {"results": results}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    config = load_config()["server"]
    logging.basicConfig(level=config["log_level"])
    uvicorn.run(app, host=config["host"], port=int(config["port"]))
